import time

from pycudwt import Wavelets
import numpy as np
from PIL import Image

from imwatermark import WatermarkEncoder

np.random.seed(42)

METHOD_DWT_DCT = 'dwtDct'
WATERMARK_MESSAGE = 0b101100111110110010010000011110111011000110011110
WATERMARK_BITS = [int(bit) for bit in bin(WATERMARK_MESSAGE)[2:]]

# warmup GPU
Wavelets(np.random.randint(low=0, high=255, size=(1024, 1024), dtype=np.uint8), "haar", 1)
WatermarkEncoder().warmup_gpu()


class TestWavelets:
    def test_fast_watermarking(self):
        """
        Ensure our GPU warmup worked correctly. Otherwise first request every time will
        take around 2 seconds to load CUDA libraries, which is too long.
        """
        start = time.time()
        wm = WatermarkEncoder()
        image_rgb = np.random.randint(low=0, high=255, size=(1024, 1024, 3), dtype=np.uint8)
        wm.set_watermark("bits", WATERMARK_BITS)
        image_bgr = np.array(image_rgb)[:, :, ::-1]
        watermarked_bgr = wm.encode(image_bgr, METHOD_DWT_DCT)
        watermarked_rgb = Image.fromarray(watermarked_bgr[:, :, ::-1])
        elapsed_ms = (time.time() - start) * 1000.

        # if it doesn't work, it will take around 2000ms, in my testing on A100s
        # if it works, should take ~50ms
        ENCODING_TOO_LONG_MS = 65
        assert elapsed_ms <= ENCODING_TOO_LONG_MS, f"Watermarking took {elapsed_ms:.2f} ms, which is too long"

    def test_pycudwt_installed_correctly(self):
        """
        If pycudwt is install incorrectly, the DWT transform will return all zeros.
        """
        # create wavelets information and plug in random data
        start = time.time()
        wv = Wavelets(np.random.randint(low=0, high=255, size=(1024, 1024), dtype=np.uint8), "haar", 1)

        # perform the discrete wavelets transform
        wv.forward()  # wv.coeffs = [A, [H1, V1, D1]]
        NON_ZERO_THRESH_SUM = 1.0
        assert np.sum(np.abs(wv.coeffs[0])) > NON_ZERO_THRESH_SUM, \
            "pycudwt is returning all zeros, did you `export PYCUDWT_CC=80` before running `pip install`?"
        
        elapsed_ms = (time.time() - start) * 1000.

        # if it works, should take <10ms! but let's put a 4x buffer for random CI machines :laugh:
        PYCUDWT_TOO_LONG_MS = 10 * 4
        assert elapsed_ms <= PYCUDWT_TOO_LONG_MS, f"pycudwt set_image() and forward() took {elapsed_ms:.2f} ms, which is too long"




import time

from pycudwt import Wavelets
import numpy as np
from PIL import Image

from imwatermark import WatermarkEncoder, WatermarkDecoder

np.random.seed(42)

METHOD_DWT_DCT = 'dwtDct'
WATERMARK_MESSAGE = 0b101100111110110010010000011110111011000110011110
WATERMARK_BITS = [int(bit) for bit in bin(WATERMARK_MESSAGE)[2:]]

# warmup GPU
Wavelets(np.random.randint(low=0, high=255, size=(1024, 1024), dtype=np.uint8), "haar", 1)
WatermarkEncoder().warmup_gpu()

def apply_watermark(image: Image.Image, method: str, bits: list) -> Image.Image:
    encoder = WatermarkEncoder()
    encoder.set_watermark("bits", bits)
    watermarked = encoder.encode(np.array(image)[:, :, ::-1], method)
    return Image.fromarray(watermarked[:, :, ::-1])

class TestEncodeDecode:
    def test_encode_decode(self):
        """
        Ensure we can encode and decode a watermark without any errors.
        """
        image = Image.open("test_vectors/original.jpg")
        image = image.crop((0, 0, 1024, 1024))

        start = time.time()
        watermarked = apply_watermark(image, METHOD_DWT_DCT, WATERMARK_BITS)
        watermarking_ms = (time.time() - start) * 1000.
        print(f"Watermarking took {watermarking_ms:.2f} ms")

        start = time.time()
        enc_length = 48
        decoder = WatermarkDecoder('bits', enc_length)
        decoding_ms = (time.time() - start) * 1000.
        print(f"Decoding took {decoding_ms:.2f} ms")



import cv2
import numpy as np
from pycudwt import Wavelets


class EmbedMaxDct(object):
    def __init__(self, watermarks=[], wmLen=8, scales=[0, 36, 36], block=4):
        self._watermarks = watermarks
        self._wmLen = wmLen
        self._scales = scales
        self._block = block

        # Create a wavelets instance - this is just to "warmup" the GPU by loading cuda libraries.
        # Note: calling it in just this instance of EmbedMaxDct will warmup for all future instances
        # of EmbedMaxDct in same Python process!
        Wavelets(
            np.random.randint(low=0, high=255, size=(1024, 1024), dtype=np.uint8),
            "haar",
            1,
        )

    def encode(self, bgr):
        (row, col, channels) = bgr.shape
        yuv = cv2.cvtColor(bgr, cv2.COLOR_BGR2YUV)

        for channel in range(2):
            if self._scales[channel] <= 0:
                continue

            # send image to GPU
            wv = Wavelets(yuv[: row // 4 * 4, : col // 4 * 4, channel], "haar", 1)

            # perform the discrete wavelets transform
            wv.forward()  # wv.coeffs = [A, [H1, V1, D1]]

            # encode our coefficients with bit sequence
            encoded_approx_matrix = self.encode_frame(
                wv.coeffs[0], self._scales[channel]
            )

            # load the encoded coefficients back into the wavelets instance in GPU memory
            # and perform inverse discrete wavelets transform
            wv.set_coeff(encoded_approx_matrix, 0, 0)
            wv.inverse()

            # load the inverse wavelets transform back into the image
            yuv[: row // 4 * 4, : col // 4 * 4, channel] = np.array(wv.image).astype(
                np.uint8
            )

        bgr_encoded = cv2.cvtColor(yuv, cv2.COLOR_YUV2BGR)
        return bgr_encoded

    def decode(self, bgr):
        (row, col, channels) = bgr.shape

        yuv = cv2.cvtColor(bgr, cv2.COLOR_BGR2YUV)

        scores = [[] for i in range(self._wmLen)]
        for channel in range(2):
            if self._scales[channel] <= 0:
                continue

            wv = Wavelets(yuv[: row // 4 * 4, : col // 4 * 4, channel], "haar", 1)

            scores = self.decode_frame(wv.coeffs[0], self._scales[channel], scores)

        avgScores = list(map(lambda l: np.array(l).mean(), scores))

        bits = np.array(avgScores) * 255 > 127
        return bits

    def decode_frame(self, frame, scale, scores):
        (row, col) = frame.shape
        num = 0

        for i in range(row // self._block):
            for j in range(col // self._block):
                block = frame[
                    i * self._block : i * self._block + self._block,
                    j * self._block : j * self._block + self._block,
                ]

                score = self.infer_dct_matrix(block, scale)
                wmBit = num % self._wmLen
                scores[wmBit].append(score)
                num = num + 1

        return scores

    def infer_dct_matrix(self, block, scale):
        pos = np.argmax(abs(block.flatten()[1:])) + 1
        i, j = pos // self._block, pos % self._block

        val = block[i][j]
        if val < 0:
            val = abs(val)

        if (val % scale) > 0.5 * scale:
            return 1
        else:
            return 0

    def encode_frame(self, frame, scale):
        """
        frame is a matrix (M, N)

        we get K (watermark bits size) blocks (self._block x self._block)

        For i-th block, we encode watermark[i] bit into it
        """
        (row, col) = frame.shape
        num_rows = row // self._block
        num_cols = col // self._block
        num_cells = num_rows * num_cols

        # generate our sequence of watermark bits to encode
        incrementing_num = np.arange(num_rows * num_cols, dtype=np.int16).reshape(
            (num_rows, num_cols)
        )
        wmBits = np.array(self._watermarks)[(incrementing_num % self._wmLen).flatten()]

        block_m = self._block  # rows in block
        block_n = self._block  # cols in block
        blocksize = block_m * block_n

        # find the non-upper left maxiumum coefficient (i, j) index per block
        blocks = frame.reshape(num_rows, self._block, num_cols, self._block).swapaxes(
            1, 2
        )
        max_block_val_idx = (
            np.argmax(np.abs(blocks.reshape(num_cells, blocksize)[:, 1:]), axis=1) + 1
        )
        max_block_i = max_block_val_idx // self._block
        max_block_j = max_block_val_idx % self._block

        # extract the maximum coefficient value per block
        block_idx = np.arange(num_cells)
        block_idx_to_block = blocks.reshape(num_cells, self._block, self._block)
        max_block_vals = block_idx_to_block[
            block_idx, max_block_i.flatten(), max_block_j.flatten()
        ]

        # perform the bit stuffing operation using the indices & values
        block_idx_to_block[block_idx, max_block_i.flatten(), max_block_j.flatten()] = (
            np.sign(max_block_vals)
            * (
                np.abs(max_block_vals) // scale
                + 0.25
                + 0.5 * wmBits[incrementing_num.flatten()].flatten()
            )
            * scale
        )

        # reshape the blocks back into the original frame shape
        encoded_frame = (
            block_idx_to_block.reshape(num_rows, num_cols, self._block, self._block)
            .transpose(0, 2, 1, 3)
            .reshape(row, col)
        )
        return encoded_frame



import numpy as np
import torch
import cv2
import os
import time


class RivaWatermark(object):
    encoder = None
    decoder = None

    def __init__(self, watermarks=[], wmLen=32, threshold=0.52):
        self._watermarks = watermarks
        self._threshold = threshold
        if wmLen not in [32]:
            raise RuntimeError('rivaGan only supports 32 bits watermarks now.')
        self._data = torch.from_numpy(np.array([self._watermarks], dtype=np.float32))

    @classmethod
    def loadModel(cls):
        try:
            import onnxruntime
        except ImportError:
            raise ImportError(
                "The `RivaWatermark` class requires onnxruntime to be installed. "
                "You can install it with pip: `pip install onnxruntime`."
            )

        if RivaWatermark.encoder and RivaWatermark.decoder:
            return
        modelDir = os.path.dirname(os.path.abspath(__file__))
        RivaWatermark.encoder = onnxruntime.InferenceSession(
            os.path.join(modelDir, 'rivagan_encoder.onnx'))
        RivaWatermark.decoder = onnxruntime.InferenceSession(
            os.path.join(modelDir, 'rivagan_decoder.onnx'))

    def encode(self, frame):
        if not RivaWatermark.encoder:
            raise RuntimeError('call loadModel method first')

        frame = torch.from_numpy(np.array([frame], dtype=np.float32)) / 127.5 - 1.0
        frame = frame.permute(3, 0, 1, 2).unsqueeze(0)

        inputs = {
            'frame': frame.detach().cpu().numpy(),
            'data': self._data.detach().cpu().numpy()
        }

        outputs = RivaWatermark.encoder.run(None, inputs)
        wm_frame = outputs[0]
        wm_frame = torch.clamp(torch.from_numpy(wm_frame), min=-1.0, max=1.0)
        wm_frame = (
            (wm_frame[0, :, 0, :, :].permute(1, 2, 0) + 1.0) * 127.5
        ).detach().cpu().numpy().astype('uint8')

        return wm_frame

    def decode(self, frame):
        if not RivaWatermark.decoder:
            raise RuntimeError('you need load model first')

        frame = torch.from_numpy(np.array([frame], dtype=np.float32)) / 127.5 - 1.0
        frame = frame.permute(3, 0, 1, 2).unsqueeze(0)
        inputs = {
            'frame': frame.detach().cpu().numpy(),
        }
        outputs = RivaWatermark.decoder.run(None, inputs)
        data = outputs[0][0]
        return np.array(data > self._threshold, dtype=np.uint8)



import setuptools


with open('README.md', 'r') as f:
  long_description = f.read()

setuptools.setup(
  name='invisible-watermark-gpu',
  version='0.2.0',
  author='Will Drevo',
  author_email='will.drevo@gmail.com',
  description='GPU adaptation for the library for creating and decoding invisible image watermarks',
  long_description=long_description,
  long_description_content_type='text/markdown',
  url='https://github.com/worldveil/invisible-watermark',
  packages=setuptools.find_packages(),
  install_requires=[
      'opencv-python>=4.1.0.25',
      'torch>=2.0.0',
      'Pillow>=6.0.0',
      'PyWavelets>=1.1.1',
      'numpy>=1.17.0',
      'pycudwt-multitarget==1.0.5',
      'pytest',
  ],
  scripts=['invisible-watermark'],
  classifiers=[
      'Programming Language :: Python :: 3',
      'License :: OSI Approved :: MIT License',
      'Operating System :: POSIX :: Linux',
  ],
  # Python 3.6 tested in Ubuntu 18.04 LTS.
  python_requires='>=3.6',
  include_package_data=True,
)



import numpy as np
import copy
import cv2
import pywt
import math
import pprint

pp = pprint.PrettyPrinter(indent=2)


class EmbedDwtDctSvd(object):
    def __init__(self, watermarks=[], wmLen=8, scales=[0,36,0], block=4):
        self._watermarks = watermarks
        self._wmLen = wmLen
        self._scales = scales
        self._block = block

    def encode(self, bgr):
        (row, col, channels) = bgr.shape

        yuv = cv2.cvtColor(bgr, cv2.COLOR_BGR2YUV)

        for channel in range(2):
            if self._scales[channel] <= 0:
                continue

            ca1,(h1,v1,d1) = pywt.dwt2(yuv[:row//4*4,:col//4*4,channel], 'haar')
            self.encode_frame(ca1, self._scales[channel])

            yuv[:row//4*4,:col//4*4,channel] = pywt.idwt2((ca1, (v1,h1,d1)), 'haar')

        bgr_encoded = cv2.cvtColor(yuv, cv2.COLOR_YUV2BGR)
        return bgr_encoded

    def decode(self, bgr):
        (row, col, channels) = bgr.shape

        yuv = cv2.cvtColor(bgr, cv2.COLOR_BGR2YUV)

        scores = [[] for i in range(self._wmLen)]
        for channel in range(2):
            if self._scales[channel] <= 0:
                continue

            ca1,(h1,v1,d1) = pywt.dwt2(yuv[:row//4*4,:col//4*4,channel], 'haar')

            scores = self.decode_frame(ca1, self._scales[channel], scores)

        avgScores = list(map(lambda l: np.array(l).mean(), scores))

        bits = (np.array(avgScores) * 255 > 127)
        return bits

    def decode_frame(self, frame, scale, scores):
        (row, col) = frame.shape
        num = 0

        for i in range(row//self._block):
            for j in range(col//self._block):
                block = frame[i*self._block : i*self._block + self._block,
                              j*self._block : j*self._block + self._block]

                score = self.infer_dct_svd(block, scale)
                wmBit = num % self._wmLen
                scores[wmBit].append(score)
                num = num + 1

        return scores

    def diffuse_dct_svd(self, block, wmBit, scale):
        u,s,v = np.linalg.svd(cv2.dct(block))

        s[0] = (s[0] // scale + 0.25 + 0.5 * wmBit) * scale
        return cv2.idct(np.dot(u, np.dot(np.diag(s), v)))

    def infer_dct_svd(self, block, scale):
        u,s,v = np.linalg.svd(cv2.dct(block))

        score = 0
        score = int ((s[0] % scale) > scale * 0.5)
        return score
        if score >= 0.5:
            return 1.0
        else:
            return 0.0

    def encode_frame(self, frame, scale):
        '''
        frame is a matrix (M, N)

        we get K (watermark bits size) blocks (self._block x self._block)

        For i-th block, we encode watermark[i] bit into it
        '''
        (row, col) = frame.shape
        num = 0
        for i in range(row//self._block):
            for j in range(col//self._block):
                block = frame[i*self._block : i*self._block + self._block,
                              j*self._block : j*self._block + self._block]
                wmBit = self._watermarks[(num % self._wmLen)]


                diffusedBlock = self.diffuse_dct_svd(block, wmBit, scale)
                frame[i*self._block : i*self._block + self._block,
                      j*self._block : j*self._block + self._block] = diffusedBlock

                num = num+1



import base64
import logging
import struct
import time
import uuid

import numpy as np

from .maxDct import EmbedMaxDct
from .dwtDctSvd import EmbedDwtDctSvd
from .rivaGan import RivaWatermark

logger = logging.getLogger(__name__)


class WatermarkEncoder(object):
    def __init__(self, content=b""):
        seq = np.array([n for n in content], dtype=np.uint8)
        self._watermarks = list(np.unpackbits(seq))
        self._wmLen = len(self._watermarks)
        self._wmType = "bytes"

    def warmup_gpu(self):
        """
        Call this function before doing multiple encodes!

        This function is to perform `cuLibraryLoadData` operations which
        can take a couple of seconds, but are a one-time cost to perform.

        Currently only supports ['dwtDct'] method of encoding.
        """
        start = time.time()
        EmbedMaxDct(self._watermarks, wmLen=self._wmLen)
        elapsed_ms = (time.time() - start) * 1000.0
        logger.info(f"GPU warmup completed in {elapsed_ms:.2f} ms")

    def set_by_ipv4(self, addr):
        bits = []
        ips = addr.split(".")
        for ip in ips:
            bits += list(np.unpackbits(np.array([ip % 255], dtype=np.uint8)))
        self._watermarks = bits
        self._wmLen = len(self._watermarks)
        self._wmType = "ipv4"
        assert self._wmLen == 32

    def set_by_uuid(self, uid):
        u = uuid.UUID(uid)
        self._wmType = "uuid"
        seq = np.array([n for n in u.bytes], dtype=np.uint8)
        self._watermarks = list(np.unpackbits(seq))
        self._wmLen = len(self._watermarks)

    def set_by_bytes(self, content):
        self._wmType = "bytes"
        seq = np.array([n for n in content], dtype=np.uint8)
        self._watermarks = list(np.unpackbits(seq))
        self._wmLen = len(self._watermarks)

    def set_by_b16(self, b16):
        content = base64.b16decode(b16)
        self.set_by_bytes(content)
        self._wmType = "b16"

    def set_by_bits(self, bits=[]):
        self._watermarks = [int(bit) % 2 for bit in bits]
        self._wmLen = len(self._watermarks)
        self._wmType = "bits"

    def set_watermark(self, wmType="bytes", content=""):
        if wmType == "ipv4":
            self.set_by_ipv4(content)
        elif wmType == "uuid":
            self.set_by_uuid(content)
        elif wmType == "bits":
            self.set_by_bits(content)
        elif wmType == "bytes":
            self.set_by_bytes(content)
        elif wmType == "b16":
            self.set_by_b16(content)
        else:
            raise NameError("%s is not supported" % wmType)

    def get_length(self):
        return self._wmLen

    def encode(self, cv2Image, method='dwtDct', **configs):
        (r, c, channels) = cv2Image.shape
        if r*c < 256*256:
            raise ValueError('image too small, should be larger than 256x256')

        if method == 'dwtDct':
            embed = EmbedMaxDct(self._watermarks, wmLen=self._wmLen, **configs)
            return embed.encode(cv2Image)
        elif method == 'dwtDctSvd':
            embed = EmbedDwtDctSvd(self._watermarks, wmLen=self._wmLen, **configs)
            return embed.encode(cv2Image)
        elif method == 'rivaGan':
            embed = RivaWatermark(self._watermarks, self._wmLen)
            return embed.encode(cv2Image)
        else:
            raise NameError('%s is not supported' % method)


class WatermarkDecoder(object):
    def __init__(self, wm_type="bytes", length=0):
        self._wmType = wm_type
        if wm_type == "ipv4":
            self._wmLen = 32
        elif wm_type == "uuid":
            self._wmLen = 128
        elif wm_type == "bytes":
            self._wmLen = length
        elif wm_type == "bits":
            self._wmLen = length
        elif wm_type == "b16":
            self._wmLen = length
        else:
            raise NameError("%s is unsupported" % wm_type)

    def reconstruct_ipv4(self, bits):
        ips = [str(ip) for ip in list(np.packbits(bits))]
        return ".".join(ips)

    def reconstruct_uuid(self, bits):
        nums = np.packbits(bits)
        bstr = b""
        for i in range(16):
            bstr += struct.pack(">B", nums[i])

        return str(uuid.UUID(bytes=bstr))

    def reconstruct_bits(self, bits):
        # return ''.join([str(b) for b in bits])
        return bits

    def reconstruct_b16(self, bits):
        bstr = self.reconstruct_bytes(bits)
        return base64.b16encode(bstr)

    def reconstruct_bytes(self, bits):
        nums = np.packbits(bits)
        bstr = b""
        for i in range(self._wmLen // 8):
            bstr += struct.pack(">B", nums[i])
        return bstr

    def reconstruct(self, bits):
        if len(bits) != self._wmLen:
            raise ValueError("bits are not matched with watermark length")

        if self._wmType == "ipv4":
            return self.reconstruct_ipv4(bits)
        elif self._wmType == "uuid":
            return self.reconstruct_uuid(bits)
        elif self._wmType == "bits":
            return self.reconstruct_bits(bits)
        elif self._wmType == "b16":
            return self.reconstruct_b16(bits)
        else:
            return self.reconstruct_bytes(bits)

    def decode(self, cv2Image, method='dwtDct', **configs):
        (r, c, channels) = cv2Image.shape
        if r*c < 256*256:
            raise ValueError('image too small, should be larger than 256x256')

        bits = []
        if method == 'dwtDct':
            embed = EmbedMaxDct(watermarks=[], wmLen=self._wmLen, **configs)
            bits = embed.decode(cv2Image)
        elif method == 'dwtDctSvd':
            embed = EmbedDwtDctSvd(watermarks=[], wmLen=self._wmLen, **configs)
            bits = embed.decode(cv2Image)
        elif method == 'rivaGan':
            embed = RivaWatermark(watermarks=[], wmLen=self._wmLen, **configs)
            bits = embed.decode(cv2Image)
        else:
            raise NameError('%s is not supported' % method)
        return self.reconstruct(bits)
